/* =========================================================================
 *
 *             Nanite Systems Applied Research Execution System
 *  
 *            Copyright (c) 2022–2023 Nanite Systems Corporation
 *  
 * =========================================================================
 *
 *  Navigation System Program
 *  
 *  This program is covered under the terms of the ARES Software Copyright
 *  License, Section 4 (ASCL-iv). It is a closed-source, proprietary,
 *  commercial product and may not be distributed.
 *  
 *  To see the full text of the ASCL, type 'help license' on any standard
 *  ARES distribution, or visit http://nanite-systems.com/ASCL for the
 *  current version.
 *
 *  DISCLAIMER
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS
 *  IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 *  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 *  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 *  DAMAGES HOWEVER CAUSED ON ANY THEORY OF LIABILITY ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGE.
 *
 * =========================================================================
 *
 */

#include <ARES/a>
#define CLIENT_VERSION ARES_VERSION
#define CLIENT_VERSION_TAGS ARES_VERSION_TAGS

// devices channel:
integer C_LIGHT_BUS;

integer teleport_enabled = TRUE;
integer location_enabled = TRUE;
integer navigation_blocked = FALSE;

string teleport_target_region;
vector teleport_target_coordinates;

key teleport_pipe;
key follow_target;

#define NAV_INTERVAL 0.5
float leash_length = 4;
float speed = 4;

integer moving;

list nav_path;
vector next_nav_dest;
vector nav_path_dest;

stop_follow() {
	task_end(follow_target);
	if(follow_target)
		deletedbl("nav", ["follow", "target"]);
	follow_target = "";
	io_tell(NULL_KEY, C_LIGHT_BUS, "follow " + (string)NULL_KEY);
}

list construct_path(key target) {
	vector origin_pos = llGetPos();
	list origin_down = llCastRay(llGetPos(), llGetPos() - <0, 0, 3>, []);
	vector target_pos = object_pos(target);
	list target_down = llCastRay(target_pos, target_pos - <0, 0, 3>, []);
	vector dest = getv(target_down, 1);
	if(dest == ZV) {
		vector target_bb = llGetAgentSize(target);
		dest = target_pos - <0, 0, target_bb.z * 0.5>;
	}
	vector origin = getv(origin_down, 1);
	if(origin == ZV) {
		vector origin_bb = llGetAgentSize(avatar);
		origin = origin_pos - <0, 0, origin_bb.z * 0.5>;
	}
	return llGetStaticPath(origin, dest, 0.4, []);
}

main(integer src, integer n, string m, key outs, key ins, key user) {
	if(n == SIGNAL_INVOKE) {
		list argv = split(m, " ");
		integer argc = count(argv);
		string action;
		string msg;
		if(argc == 1) {
			msg = "Usage:\n    "
			+ PROGRAM_NAME + " follow [<key>]: follows the specified object\n    "
			+ PROGRAM_NAME + " bookmark <name> remove|here|<slurl>: manages a TP bookmark\n    "
			+ PROGRAM_NAME + " add <name> here|<slurl>: as bookmark\n    "
			+ PROGRAM_NAME + " remove <name>: remove a bookmark\n    "
			+ PROGRAM_NAME + " tp|warp|jump <name>|<slurl>|me: TPs to a bookmark or slurl\n    "
			+ PROGRAM_NAME + " end|stop: stops current navigation\n    "
			+ PROGRAM_NAME + " start <dest>: begins navigating to the specified destination (not yet implemented)";
		;
		} else if((action = gets(argv, 1)) == "start") {
			navigation_blocked = (integer)getdbl("nav", ["blocked"]);
			if(navigation_blocked || !location_enabled) {
				if(task_count()) {
					task_end(NULL_KEY);
					msg = "Aborted navigation due to ";
				} else {
					msg = "Cannot navigate due to ";
				}
				
				if(!location_enabled) {
					msg += " location sensor anomalies.";
				} else {
					msg += " electromagnetic interference.";
				}
			
			} else if(task_count()) {
				msg = "Already running.";
			} else {
				task_begin(NULL_KEY, "nav");
				msg = "Started navigation.";
				llSetTimerEvent(NAV_INTERVAL);
			}
		} else if(action == "end" || action == "stop") {
			if(task_count()) {
				msg = "???? current tasks are " + concat(jskeys(tasks_queue), ", ");
				
				if(getjs(tasks_queue, [NULL_KEY]) != JSON_INVALID) {
					task_end(NULL_KEY);
					msg = "Stopped navigation.";
				}
				
				if(follow_target) {
					stop_follow();
					msg = "Stopped following.";
				}
				
			} else {
				msg = "Not running.";
			}
		} else if(action == "path") {
			key target = gets(argv, 2);
			if(target) {} else target = user;
			msg = concat(construct_path(target), "\n");
		} else if(action == "refresh") {
			string topic = gets(argv, 2);
			if(topic == "teleport") {
				list names = jskeys(getdbl("nav", ["bookmark"]));
				list buttons = [];
				integer bi = count(names);
				while(bi--) {
					string name = gets(names, bi);
					string b = jsarray([
						name,
						0,
						"nav teleport " + name
					]);
					buttons = b + buttons;
				}
				setdbl("m:teleport", ["d"], jsarray(buttons));
			}
		} else if(action == "status") {
			msg = "Current tasks: " + tasks_queue + " = " + (string)task_count();
		} else if(action == "follow") {
			if(task_count()) {
				msg = "Already navigating. Run 'nav stop' to abort.";
			} else {
				key target = gets(argv, 2);
				if(target == "")
					target = user;
				if(target != NULL_KEY && target != avatar) {
					msg = "Following " + llKey2Name(target) + ".";
					leash_length = (float)getdbl("nav", ["follow", "range"]);
					speed = (float)getdbl("nav", ["speed"]);
					if(speed == 0) speed = 4;
					if(leash_length == 0) leash_length = 2;
					setdbl("nav", ["follow", "target"], target);
					io_tell(NULL_KEY, C_LIGHT_BUS, "follow " + (string)target);
					task_begin(follow_target = target, "follow");
					llSetTimerEvent(NAV_INTERVAL);
				} else {
					msg = "Cannot follow nothing.";
				}
			}
		} else if(action == "bookmark" || action == "add" || action == "remove") {
			string name = gets(argv, 2);
			string location;
			if(action == "remove")
				location = action;
			else
				location = concat(delrange(argv, 0, 2), " ");
			
			if(location == "here" || location == "") {
				if(!location_enabled) {
					msg = "Cannot get current position while the location subsystem is disabled.";
					jump fail;
				}
				vector p = llGetPos();
				location = llGetRegionName()
					+ "/" + (string)((integer)llRound(p.x))
					+ "/" + (string)((integer)llRound(p.y))
					+ "/" + (string)((integer)llRound(p.z));
			} else if(location == "remove") {
				location = getdbl("nav", ["bookmark", name]);
				if(location != JSON_INVALID) {
					deletedbl("nav", ["bookmark", name]);
					msg = "Removed bookmark " + name;
				} else {
					msg = "No bookmark to remove: " + name;
				}
				jump fail;
			} else if(substr(location, 0, 37) == "http://maps.secondlife.com/secondlife/") {
				location = delstring(location, 0, 37);
			} else {
				msg = "Unknown SLURL format: " + location + ". Please specify a SLURL starting with http://maps.secondlife.com/secondlife/.";
				jump fail;
			}
			setdbl("nav", ["bookmark", name], location);
			msg = "Bookmark '" + name + "' set.";
		} else if(action == "tp" || action == "warp" || action == "jump" || action == "teleport") {
			string dest = gets(argv, 2);
			string location = getdbl("nav", ["bookmark", dest]);
			
			if(location != JSON_INVALID)
				dest = "http://maps.secondlife.com/secondlife/" + location;
			
			if(!location_enabled) {
				msg = "Cannot teleport safely while the location subsystem is disabled.";
				jump fail;
			} else if(substr(dest, 0, 37) == "http://maps.secondlife.com/secondlife/") {
				list parts = splitnulls(dest, "/");
				teleport_target_region = gets(parts, 4);
				teleport_target_coordinates = <
					(integer)gets(parts, 5),
					(integer)gets(parts, 6),
					(integer)gets(parts, 7)
				>;
			} else if(dest == "me") {
				teleport_target_region = llGetRegionName();
				teleport_target_coordinates = object_pos(user);
			} else {
				msg = "No bookmark: " + dest;
				jump fail;
			}
			
			teleport_pipe = llGenerateKey();
			pipe_open(["p:" + (string)teleport_pipe + " notify " + PROGRAM_NAME + " region"]);
		}
		@fail;
		
		if(msg != "")
			print(outs, user, msg);
	} else if(n == SIGNAL_NOTIFY) {
		list argv = split(m, " ");
		integer argc = count(argv);
		string msg;
		integer argi = 1;
		while(argi < argc) {
			string action = gets(argv, argi);
			string val = gets(argv, argi + 1);
			
			if(action == "region") {
				string buffer;
				pipe_read(ins, buffer);
				
				list parts = splitnulls(buffer, " = ");
				vector region_pos = (vector)gets(parts, LAST);
				vector tp_dest = region_pos + teleport_target_coordinates;
				if(teleport_enabled) {
					echo("@tpto:" + (string)tp_dest.x
						+ "/" + (string)tp_dest.y
						+ "/" + (string)tp_dest.z + "=force");
				} else {
					echo("Teleport failed: subsystem is not online.");
				}
				
				pipe_close([ins]);
			} else if(action == "pipe") {
				request_regionpos(teleport_target_region, ins, avatar);
			} else if(action == "abort") {
				navigation_blocked = (integer)getdbl("nav", ["blocked"]);
				if(task_count() && navigation_blocked) {
					task_end(NULL_KEY);
					msg = "Aborted navigation due to electromagnetic interference.";
				}
			} else if(action == "teleport") {
				teleport_enabled = (val == "y");
			} else if(action == "location") {
				location_enabled = (val == "y");
				if(task_count() && !location_enabled) {
					task_end(NULL_KEY);
					msg = "Aborted navigation due to location sensor anomalies.";
				}
			}
			
			if(msg != "") {
				print(outs, user, msg);
				msg = "";
			}
			argi += 2;
		}
	} else if(n == SIGNAL_INIT) {
		#ifdef DEBUG
			echo("[" + PROGRAM_NAME + "] init event");
		#endif
		navigation_blocked = (integer)getdbl("nav", ["blocked"]);
		notify_program("_power notify " + PROGRAM_NAME, avatar, NULL_KEY, avatar);
		
		llStopMoveToTarget();
		
		if(llGetAttached())
			C_LIGHT_BUS = 105 - (integer)("0x" + substr(avatar, 29, 35));
		else
			C_LIGHT_BUS = 105 - (integer)("0x" + substr(KERNEL, 29, 35));
	} else {
		echo("[" + PROGRAM_NAME + "] unimplemented signal " + (string)n + ": " + m);
	}
}

#define EXT_EVENT_HANDLER "ARES/application/nav.event.lsl"
#include <ARES/program>
