/* =========================================================================
 *
 *             Nanite Systems Applied Research Execution System
 *  
 *            Copyright (c) 2022–2023 Nanite Systems Corporation
 *  
 * =========================================================================
 *
 *  Input System Module
 *
 *  This program is covered under the terms of the ARES Software Copyright
 *  License, Section 1 (ASCL-i). It is offered to you on a limited basis to
 *  facilitate modification and customization.
 *  
 *  To see the full text of the ASCL, type 'help license' on any standard
 *  ARES distribution, or visit http://nanite-systems.com/ASCL for the
 *  current version.
 *
 *  DISCLAIMER
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS
 *  IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 *  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 *  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 *  DAMAGES HOWEVER CAUSED ON ANY THEORY OF LIABILITY ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGE.
 *
 * =========================================================================
 *
 */

#include <ARES/a>
#define CLIENT_VERSION ARES_VERSION
#define CLIENT_VERSION_TAGS ARES_VERSION_TAGS

// it's good enough for real computers:
#define NON_VOLATILE
#define DAEMON_LISTEN

key input_handle;
integer user_channel = 7770;

integer voice_state = 1; // subsystem: allow any talking at all
integer mind_state = 1; // subsystem: allow original speaking
integer speech_state = 1; // subsystem: allow say
integer amplifier_state = 1; // subsystem: allow shouting
integer hearing_state = 1; // subsystem: allow hearing

integer allow_recvchat = 1; // allowed to receive chat?
integer allow_recvemote = 1; // allowed to receive emotes?
integer allow_sendchat = 1; // allowed to send chat?

integer hearing_captured;

float impedence_in;
float impedence_out;
float impedence_mind;

integer vox_output = FALSE;
integer vox_input = FALSE;

string vo_intake;
list vo_cells; // [cell-split message currently being processed]
list vo_queue; // [messages to process, left as strings]
integer vo_next_cell = NOWHERE;
string output_pipe = OUTPUT_SAY;

string vi_intake;
string vi_source;
list vi_cells; // [cell-split message currently being processed]
list vi_queue; // [source, message]
integer vi_next_cell = NOWHERE;

#define VO_ADD 0x01
#define VO_ALTER 0x03
// #define VO_PASS 0x05
#define VI_ADD 0x08
#define VI_ALTER 0x10
// #define VI_PASS 0x14

string vi_name(string vis) {
	string visn;
	if(llGetAgentSize(vis) != ZV) {
		visn = llGetDisplayName(vis);
	} else {
		visn = llKey2Name(vis);
	}
	return visn;
}

string file_name;
integer file_offset;
integer file_length;
integer file_type; // 0 = lines, 1 = chars
key file_pipe;
key file_outs;
key file_user;
string script;
list script_lines;
integer script_li;
key script_handle = NULL_KEY; // don't continue executing a script until we get a suitable DONE back

string expression(list tokens) {
	if((string)tokens == "undefined")
		return JSON_INVALID;
	
	float value = (float)gets(tokens, 0);
	integer ki = 1;
	integer kmax = count(tokens);
	while(ki < kmax) {
		string op = gets(tokens, ki);
		float r = (float)gets(tokens, ki + 1);
		if(op == "!=") {
			value = (float)(llFabs(value - r) >= 0.001);
		} else if(op == "==") {
			value = (float)(llFabs(value - r) < 0.001);
		} else if(op == ">") {
			value = (float)((value - r) > 0.001);
		} else if(op == "<") {
			value = (float)((r - value) > 0.001);
		} else if(op == ">=") {
			value = (float)((value - r) > -0.001);
		} else if(op == "<=") {
			value = (float)((r - value) > -0.001);
		} else if(op == "+") {
			value += r;
		} else if(op == "-") {
			value -= r;
		} else if(op == "*") {
			value *= r;
		} else if(op == "/") {
			if(r != 0) {
				value /= r;
			} else {
				echo("[_input] division by 0 in expression: " + concat(tokens, " "));
				return "0";
			}
		} else {
			echo("[_input] unknown operator " + op + " in expression: " + concat(tokens, " "));
			return "0";
		}
		ki += 2;
	}
	
	// return integer if within a thousandth:
	if(llFabs((float)((integer)value) - value) < 0.001)
		return (string)((integer)value);
	else
		return (string)value;
}

#define play_speech_marker() e_call(C_EFFECTOR, E_SIGNAL_CALL, (string)avatar + " " + (string)avatar + " effector marker")

integer implicit_at = FALSE; // on: expect " for chat; off: expect @ for commands (/ always works for chat)

process_input(key outs, key user, string cml, integer in_script, integer mind_required) {
	@internal_restart;
	key handle = NULL_KEY;
	if(in_script)
		script_handle = handle = llGenerateKey();
	
	// echo("processing " + cml);
	string first = substr(cml, 0, 0);
	string second = substr(cml, 0, 1);
	
	if(~strpos(cml, "\\n"))
		cml = replace(cml, "\\n", "\n");
	
	// =======================================
	// shell control commands
	// =======================================
	
	if(first == "=") {
		list cargv = splitnulls(cml, " ");
		string prog = gets(cargv, 0);
		if(prog == "=pipes" || prog == "=purge") {
			integer purge = (prog == "=purge");
			list pipes = llLinksetDataFindKeys("p:", 0, 0);
			integer pi = count(pipes);
			if(pi == 0)
				echo("No pipes have waiting data.");
			else {
				if(purge)
					echo("Pipes to purge: " + (string)pi);
				
				while(pi--) {
					string pn = gets(pipes, pi);
					echo(delstring(pn, 0, 1) + " = " + llLinksetDataRead(pn));
					if(purge)
						llLinksetDataDelete(pn);
				}
			}
		} else if(prog == "=ddt") {
			linked(LINK_ROOT, 0x1000, gets(cargv, 1), user);
		} else if(prog == "=encode") {
			echo("ARES encode: " + ares_encode((integer)gets(cargv, 1)));
		} else if(prog == "=decode") {
			string d = gets(cargv, 1);
			integer da = ares_decode(d);
			echo("ARES decode: " + (string)da + " / hex 0x" + hex(da));
		} else if(prog == "=alias") {
			string word = gets(cargv, 1);
			string command = concat(delrange(cargv, 0, 1), " ");
			if(word == "delete" && command != "") {
				string alias = getdbl("input", ["alias", command]);
				if(alias != JSON_INVALID)
					setdbl("input", ["alias", command], JSON_DELETE);
				echo("OK.");
			} else if(command != "") {
				setdbl("input", ["alias", word], command);
			} else if(word != "") {
				echo(getdbl("input", ["alias", word]));
			} else {
				list aliases = js2list(getdbl("input", ["alias"]));
				integer ai = count(aliases) >> 1;
				if(!ai)
					echo("No aliases defined.");
				else while(ai--) {
					echo(" - " + gets(aliases, ai << 1) + ": " + gets(aliases, (ai << 1) + 1));
				}
			}
		} else if(prog == "=terminal") {
			if(!implicit_at)
				print(outs, user, "Enabled terminal mode.");
			implicit_at = TRUE;
		} else if(prog == "=chat") {
			if(!implicit_at)
				print(outs, user, "Enabled chat mode.");
			implicit_at = FALSE;
		} else if(prog == "=alert") {
			e_call(C_VARIATYPE, E_SIGNAL_CALL, (string)outs + " " + (string)user + " variatype alert response " + gets(cargv, 1));
		} else {
			llSetMemoryLimit(65535);
			llSetMemoryLimit(65536);
			echo("? memory: " + (string)llGetFreeMemory());
		}
	} else if(first == "." && llToUpper(second) != llToLower(second)) {
		invoke("persona " + cml, outs, handle, user);
	} else if(first == "!" && llToUpper(second) != llToLower(second)) {
		e_call(C_EFFECTOR, E_SIGNAL_CALL, (string)outs + " " + (string)user + " effector " + cml);
	
	// =======================================
	// system commands
	// =======================================
	
	} else if(first == "@"
		|| in_script
		|| (implicit_at
			&& first != "/"
			&& first != "\""
			&& first != ":"
			&& second != "w "
			&& second != "s "
			)) {
		
		if(!(implicit_at || in_script) || first == "@")
			cml = strdelete(cml, 0, 0);
		
		integer varpos = strpos(cml, "$");
		
		if(~varpos) {
			integer vi = varpos;
			integer vL = strlen(cml);
			string varacc;
			integer vaL = 0;
			integer vaStart;
			integer vp_state = 0;
			while(vi < vL) {
				integer vc = llOrd(cml, vi);
				++vi;
				if(vc == 0x24 && vp_state == 0) { // '$'
					if(llOrd(cml, vi - 2) == 0x5c) { // '\'
						cml = delstring(cml, vi - 2, vi - 2);
						--vi;
						--vL;
					} else {
						varacc = "";
						vaL = 0;
						vaStart = vi - 1;
						vp_state = 1;
					}
				} else if(vp_state == 1) {
					integer vd = llOrd(cml, vi);
					//if(vc == 0x2e) echo((string)vd);
					if((vc >= 0x41 && vc <= 0x5a) // uppercase
					   || (vc >= 0x61 && vc <= 0x7a) // lowercase
					   || vc == 0x5f || vc == 0x2d // _ and -
					   || (vaL > 0 && vc == 0x2e && (vd >= 0x61 && vd <= 0x7a)) // . followed by lowercase letter
					   || (vaL > 0 && vc == 0x2e && (vd >= 0x30 && vd <= 0x39)) // . followed by digit
					   || (vaL > 0 && (vc >= 0x30 && vc <= 0x39))) { // digits if non-first
						varacc += llChar(vc);
						// echo("added " + (string)vc);
						++vaL;
					} else {
						vp_state = 0;
					}
					
					if(vi == vL && vp_state) {
						vp_state = 0; // end automatically at string end
						++vi;
					}
				}
				
				if(vaL && vp_state == 0) {
					// echo("subbing in " + varacc);
					string sub;
					if(varacc == "name") {
						sub = llGetDisplayName(user);
					} else if(varacc == "user") {
						sub = (string)user;
					} else if(varacc == "self") {
						sub = (string)avatar;
					} else if(varacc == "me") {
						sub = getdbl("id", ["callsign"]);
					} else {
						sub = getdbl("env", split(varacc, "."));
					}
					cml = delstring(cml, vaStart, vi - 2);
					cml = llInsertString(cml, vaStart, sub);
					vi -= vaL;
					vi += strlen(sub) - 2;
					vL = strlen(cml);
					vaStart = 0;
					vaL = 0;
					varacc = "";
				}
			}
		}
	
		list cargv = splitnulls(cml, " ");
		string prog = gets(cargv, 0);
		string aliascmd;
		
		string error_msg;
		
		if((aliascmd = getdbl("input", ["alias", prog])) != JSON_INVALID) {
			cml = aliascmd + " " + concat(delitem(cargv, 0), " ");
			cargv = splitnulls(cml, " ");
			prog = gets(cargv, 0);
		}
		
		if(prog == "to") {
			string statement = concat(delrange(cargv, 0, 1), " ");
			if(!(in_script || implicit_at))
				statement = "@" + statement;
			
			outs = gets(cargv, 1);
			cml = statement;
			jump internal_restart;
			// process_input(gets(cargv, 1), user, statement, in_script);
		} else if(prog == "from") {
			handle = gets(cargv, 1);
			if(in_script)
				script_handle = handle;
			invoke(concat(delrange(cargv, 0, 1), " "), outs, handle, user);
		} else if(prog == "do") {
			setdbl("env", ["arg"], jsarray(delrange(cargv, 0, 0)));
			setdbl("env", ["args"], (string)(count(cargv) - 2));
		
			// if called with '@input do', these will be overwritten immediately:
			script_trigger_ins = NULL_KEY;
			script_trigger_resolve = 0;
		
			file_offset = file_length = NOWHERE;
			file_outs = outs;
			file_user = user;
			file_open(file_pipe = llGenerateKey(), file_name = gets(cargv, 1));
			// task_begin(file_pipe, file_name);
			
			// not checking in_script; allow manual interruption
			if(script_lines != []) {
				llSetTimerEvent(0);
				script_lines = [];
				script_li = 0;
				script_handle = NULL_KEY;
				set_mode(_mode & ~MODE_ACCEPT_DONE);
				resolvec(script_trigger_resolve, script_trigger_ins);
				echo("[_input] script interrupted by new @do");
			}
			return; // do not continue_script();
		} else if(prog == "exit") {
			// not checking in_script; allow manual interruption
			if(script_lines != []) {
				script_lines = [];
				script_li = 0;
				script_handle = NULL_KEY;
				set_mode(_mode & ~MODE_ACCEPT_DONE);
				resolvec(script_trigger_resolve, script_trigger_ins);
				llSetTimerEvent(0);
			}
			setdbl("env", ["arg"], "[]");
			setdbl("env", ["args"], "0");
			return; // do not continue_script();
		} else if(prog == "say") {
			cml = concat(delitem(cargv, 0), " ");
			jump internal_restart;
			// process_input(outs, user, concat(delitem(cargv, 0), " "), in_script);
		} else if(prog == "jump") {
			integer new_script_li = index(script_lines, "@" + gets(cargv, 1) + ":");
			if(~new_script_li) {
				script_li = new_script_li + 1;
			} else {
				error_msg = "[_input] script error: could not jump to " + gets(cargv, 1);
			}
		} else if(prog == "echo") {
			print(outs, user, replace(concat(delitem(cargv, 0), " "), "\\n", "\n"));
		} else if(prog == "set") {
			if(count(cargv) > 2) {
				string value;
				if(gets(cargv, 2) == "=")
					value = expression(delrange(cargv, 0, 2));
				else
					value = concat(delrange(cargv, 0, 1), " ");
				
				string sv = gets(cargv, 2);
				
				if(value == "%key")
					value = llGenerateKey();
				else if(value == "%undefined")
					value = JSON_DELETE;
				else if(sv == "%count") {
					string mode = gets(cargv, 3);
					value = concat(delrange(cargv, 0, 3), " ");
					if(mode == "words") {
						value = (string)count(split(value, " "));
					} else if(mode == "lines") {
						value = (string)count(splitnulls(value, "\n"));
					} else if(mode == "chars") {
						value = (string)strlen(value);
					}
				} else if((gets(cargv, 4) == "in") && (sv == "%word" || sv == "%line" || sv == "%char")) {
					// set <var> %word $i in $text
					integer i = (integer)gets(cargv, 3);
					value = concat(delrange(cargv, 0, 4), " ");
					// echo("command: " + cml);
					// echo("extracting " + sv + " #" + (string)i + " from " + value);
					if(sv == "%word") {
						value = gets(split(value, " "), i);
					} else if(sv == "%line") {
						value = gets(splitnulls(value, "\n"), i);
					} else if(sv == "%char") {
						value = substr(value, i, i);
					}
					// echo("got " + value);
				}
				
				setdb("env", gets(cargv, 1), value);
			} else {
				error_msg = "not enough arguments: set";
			}
		} else if(prog == "if") {
			integer endpoint = index(cargv, "then");
			if(~endpoint) {
				string statement = concat(delrange(cargv, 0, endpoint), " ");
				if(!(in_script || implicit_at))
					statement = "@" + statement;
				
				integer success = 0;
				if(gets(cargv, 2) == "is") {
					success = (gets(cargv, 1) == gets(cargv, 3));
				} else {
					string value = expression(sublist(cargv, 1, endpoint - 1));
					success = (llFabs((float)value) >= 0.001);
				}
				
				if(success) {
					cml = statement;
					jump internal_restart;
					
					/*process_input(outs, user, statement, in_script);
					return; // do not continue_script(); - the true branch will handle that
					*/
				}
			} else {
				error_msg = "[_input] script error: no 'then' in 'if' statement: " + cml;
			}
		} else if(prog == "service") {
			system(SIGNAL_CALL, E_UNKNOWN + E_PROGRAM_NUMBER + (string)outs + " " + (string)user + " " + concat(delitem(cargv, 0), " "));
		} else if(~index(js2list(getdbl("input", ["device-command"])), prog)) {
			system(SIGNAL_CALL, E_HARDWARE + E_PROGRAM_NUMBER + (string)outs + " " + (string)user + " hardware command " + cml);
		} else {
			invoke(cml, outs, handle, user);
		
			return; // do not continue_script();
		}
		
		if(error_msg)
			print(outs, user, error_msg);
	
	// =======================================
	// chat processing
	// =======================================
		
	} else {
		// key destination_key = OUTPUT_SAY;
		
		if(second == "w ") {
			cml = strdelete(cml, 0, 1);
			output_pipe = OUTPUT_WHISPER;
		} else if(second == "s ") {
			cml = strdelete(cml, 0, 1);
			if(amplifier_state)
				output_pipe = OUTPUT_SHOUT;
			else
				output_pipe = OUTPUT_SAY;
		} else {
			output_pipe = OUTPUT_SAY;
		}
		
		if(!speech_state)
			output_pipe = OUTPUT_WHISPER;
		
		if(implicit_at && llOrd(cml, 0) == 0x22) // '"'
			cml = strdelete(cml, 0, 0);
		else if(llOrd(cml, 0) == 0x3a) { // ':'
			if(llOrd(cml, 1) == 0x27) // '\''
				cml = "/me" + strdelete(cml, 0, 0);
			else
				cml = "/me " + strdelete(cml, 0, 0);
		}
		
		integer slash = (llOrd(cml, 0) == 0x2f); // '/'
		integer contains_quotes = ~strpos(cml, "\"");
		integer censor = (!voice_state || (impedence_out > 1)) | 2 * (mind_required && (!mind_state || impedence_mind > 0.5));
		
		if(censor && ((slash && contains_quotes) || !slash)) {
			if(slash && contains_quotes) {
				// (this line contains immense power)
				cml = concat(llList2ListStrided(splitnulls(cml, "\""), 0, LAST, 2), "\"...\"");
			} else if(!slash) {
				cml = "/me tries to speak.";
			}
			
			if(censor & 1) {
				echo("You cannot speak.");
			} else {
				echo("You cannot think.");
			}
		}
		
		if(vox_output)
			// run_vox_queue(VO_ADD, cml, "");
			main(PROGRAM_NUMBER, SIGNAL_VOX, cml, output_pipe, NULL_KEY, avatar);
		else {
			print(output_pipe, user, cml);
			
			if((contains_quotes && voice_state) || !slash)
				play_speech_marker();
		}
	}
	
	// reduces interference from user acting during script execution
	if(in_script)
		llSetTimerEvent(0.1); // continue_script()
	/*
		TODO
		----
		block chat when unable to speak
		handle persona messages when mind is disabled
		ooc flag, "ooc", and "(("
	*/
}

integer script_trigger_resolve;
key script_trigger_ins;

continue_script() {
	integer script_length = count(script_lines);
	if(script_li < script_length) {
		string line;
		integer fc;
		while(script_li < script_length && (line == "" || fc == 0x40 || fc == 0x23)) {
			// skip blank lines and lines starting with '@' or '#'
			// echo((string)script_li + "/" + (string)script_length + " skipped");
			line = llStringTrim(gets(script_lines, script_li++), STRING_TRIM);
			fc = llOrd(line, 0);
		}
		
		// echo((string)script_li + "/" + (string)script_length + ": " + line);
		
		if(script_li <= script_length) {
			process_input(file_outs, file_user, line, TRUE, FALSE);
			return;
		}
	}
	
	#ifdef DEBUG
	echo("[_input] script done");
	#endif
	setdbl("env", ["arg"], "[]");
	setdbl("env", ["args"], "0");
	script_lines = [];
	script_li = 0;
	script_handle = NULL_KEY;
	llSetTimerEvent(0);
	set_mode(_mode & ~MODE_ACCEPT_DONE);
	
	resolvec(script_trigger_resolve, script_trigger_ins);
	
	script_trigger_resolve = 0;
}

main(integer src, integer n, string m, key outs, key ins, key user) {
	integer rvq_flag;
	string rvq_message;
	string rvq_source;

	if(n == SIGNAL_INIT) {
		string input_db = llLinksetDataRead("input");
		string shell = getjs(input_db, ["shell"]);
		
		if(user_channel) {
			// echo("@redirchat:" + (string)(user_channel) + "=y,rediremote:" + (string)(user_channel) + "=y,sendchat=y");
			effector_release("input_handle");
		}
		
		user_channel = (integer)getjs(input_db, ["channel"]);
		if(shell == JSON_INVALID)
			setdbl("input", ["shell"], (shell = "input"));
		
		if(user_channel == 0)
			user_channel = 1000 + (integer)llFrand(999);
		
		// task_begin(INPUT_HANDLER, "");
		pipe_open([
			 "p:" + INPUT_HANDLER +
			" n:" + (string)avatar +
			" from " + (string)user_channel	+
			" " + (string)avatar +
			" " + shell
		]);
		
		string vox_db = llLinksetDataRead("vox");
		
		vox_output = (integer)getjs(vox_db, ["output", "enabled"]);
		vox_input = (integer)getjs(vox_db, ["input", "enabled"]);
		
		if(vox_output)
			vo_intake = getjs(vox_db, ["output", "start"]);
		
		if(vox_input)
			vi_intake = getjs(vox_db, ["input", "start"]);
		
		// todo: maintain input listener
		
		notify_program("_power notify " + PROGRAM_NAME, avatar, NULL_KEY, avatar);
		
		hook_events([EVENT_ON_REZ]);
		
	} else if(n == SIGNAL_EVENT) {
		integer e = (integer)m;
		if(e == EVENT_ON_REZ) {
			if(user_channel) {
				effector_restrict("input_handle", "redirchat:" + (string)(user_channel) + "=?,rediremote:" + (string)(user_channel) + "=?,sendchat=?");
				llSleep(1);
				#ifdef DEBUG
				echo("[_input] redirect reestablished");
				#endif
			} else {
				llResetScript();
			}
		#ifdef DEBUG
		} else {
			echo("[_input] unexpected event: " + m);
		#endif
		}
	} else if(n == SIGNAL_DONE) {
		// echo("[_input] receipt: " + m);
		ins = substr(m, 0, 35);
		if(_mode & MODE_ACCEPT_DONE && ins == script_handle) {
			continue_script();
		} else {
			echo("[_input] expected handle " + (string)script_handle + " but got " + (string)ins);
		}
		
	} else if(n == SIGNAL_INVOKE) {
		list argv = splitnulls(llStringTrim(m, STRING_TRIM), " ");
		
		integer argc = count(argv);
		if(argc > 1) {
			string action = gets(argv, 1);
			if(action == "reset") {
				main(0, SIGNAL_INIT, "", "", "", "");
			} else if(action == "do") {
				process_input(outs, user, "@do " + concat(delrange(argv, 0, 1), " "), FALSE, FALSE);
				script_trigger_resolve = src;
				script_trigger_ins = ins;
				src = 0;
			} else if(action == "say") {
				process_input(outs, user, concat(delrange(argv, 0, 1), " "), FALSE, FALSE);
			} else {
				print(outs, user, "usage: " + gets(argv, 0) + " reset|do <filename>|say <message>");
			}
			
			if(src) {
				resolvec(src, ins);
			}
		} else if(ins == VOX_LISTENER) {
			pipe_read(ins, rvq_message);
			
			if(llOrd(rvq_message, 0) == 0x0a)
				rvq_message = delstring(rvq_message, 0, 0);
			
			#ifdef RVQ_DEBUG
			echo("incoming message: " + (string)rvq_message);
			#endif
			rvq_flag = VI_ADD;
			rvq_source = user;
			jump rvq;
		} else { // process input line
			string cm;
			pipe_read(ins, cm);
			list cms = split(cm, "\n");
			integer cmi = 0;
			integer cmimax = count(cms);
			while(cmi < cmimax) {
				string cml = llStringTrim(gets(cms, cmi), STRING_TRIM);
				if(getdbl("status", ["on"]) == "0" && ins == INPUT_HANDLER) {
					if(cml == "@on" || cml == "@boot") {
						invoke("_power on", avatar, NULL_KEY, avatar);
					} else {
						echo("System is halted. Type '@on' to boot.");
					}
				} else {
				//concat(delitem(argv, 0), ""), STRING_TRIM);
					process_input(outs, user, cml, FALSE, TRUE); 
				}
				
				++cmi;
			}
		}
	} else if(n == SIGNAL_NOTIFY) {
		// echo("INPUT notify: " + m);
		list argv = delitem(splitnulls(m, " "), 0);
		integer cimax = count(argv);
		integer ci;
		if(ci % 2) ++ci;
		while(ci < cimax) {
			string action = gets(argv, ci);
			string val = gets(argv, ci + 1);
			// echo("INPUT notify: " + action + " " + val);
			if(action == "vox-disable") {
				if(val == "output") {
					vox_output = FALSE;
					if(vo_cells != []) {
						vo_queue = [concat(vo_cells, "\"")] + vo_queue;
						integer voimax = count(vo_queue);
						integer voi = 0;
						while(voi < voimax)
							print(output_pipe, avatar, gets(vo_queue, voi++));
						
						vo_next_cell = NOWHERE;
						vo_cells = vo_queue = [];
					}
					
				} else if(val == "input") {
					vox_input = FALSE;
					
					if(vi_cells != []) {
						vi_queue = [vi_source, concat(vi_cells, "\"")] + vi_queue;
						integer viimax = count(vi_queue);
						integer vii = 0;
						while(vii < viimax) {
							llSetObjectName(vi_name(gets(vi_queue, vii)));
							echo(gets(vo_queue, vii + 1));
							vii += 2;
						}
						llSetObjectName("");
						
						vi_next_cell = NOWHERE;
						vi_cells = vi_queue = [];
					}
				}
			} else if(action == "vox-enable") {
				if(val == "output") {
					vox_output = TRUE;
					vo_intake = getdbl("vox", ["output", "start"]);
				} else if(val == "input") {
					vox_input = TRUE;
					vi_intake = getdbl("vox", ["input", "start"]);
					
					// (the input listener will be created once recvchat is blocked by the vox control program)
				}
				
			} else if(action == "output" || action == "input") { // finished processing a vox cell
				pipe_read(ins, rvq_message);
				if(action == "output")
					rvq_flag = VO_ALTER;
				else
					rvq_flag = VI_ALTER;
				
				// run_vox_queue(flag, message, "");
				jump rvq;
			} else if(action == "input-impedence") {
				impedence_in = (float)val;
			} else if(action == "output-impedence") {
				impedence_out = (float)val;
			} else if(action == "mind-impedence") {
				impedence_mind = (float)val;
			} else if(action == "voice") { // state of voice subsystem
				voice_state = (val == "y");
			} else if(action == "mind") { // state of mind subsystem
				mind_state = (val == "y");
			} else if(action == "speech") { // state of speech (base amplifier) subsystem
				speech_state = (val == "y");
			} else if(action == "amplifier") { // state of amplifier subsystem
				amplifier_state = (val == "y");
			} else if(action == "hearing") { // state of hearing subsystem
				hearing_state = (val == "y");
			} else if(action == "allow-recvchat") {
				allow_recvchat = (val == "y");
			} else if(action == "allow-recvemote") {
				allow_recvemote = (val == "y");
			} else if(action == "allow-sendchat") {
				allow_sendchat = (val == "y");
			} else if(action == "file") { // file data available
				if(ins == file_pipe) {
					string file_buffer;
					pipe_read(file_pipe, file_buffer);
					if(file_length != NOWHERE) {
						file_offset += FILE_PAGE_LENGTH;
						// print(file_outs, file_user, file_buffer);
						if(script != "" && file_type == 0)
							script += "\n";
						script += file_buffer;
						
						if(file_offset < file_length)
							file_read(file_pipe, file_name, file_offset);
						else {
							file_close(file_pipe);
							
							set_mode(MODE_MASK_BATCH);
							
							script_lines = split(script, "\n");
							script = "";
							script_li = 0;
							continue_script();
							
							// start run script here
							
							// task_end(file_pipe);
						}
					} else {
						list bfs = splitnulls(file_buffer, " ");
						file_length = (integer)gets(bfs, 0);
						file_type = (gets(bfs, 1) == "char"); // 0 = line, 1 = char
						if(file_length > 0) {
							// echo("File " + file_name + " length: " + (string)file_length);
							file_read(file_pipe, file_name, file_offset = 0);
						} else { // file not found or file empty
							// print(file_user, file_user, "No file: " + file_name);
							file_close(file_pipe);
							// task_end(file_pipe);
						}
					}
				} /*else {
					echo("File data offered via unexpected pipe: " + (string)ins);
				}*/
			} else if(action == "pipe") {
				key newk = ins;
				if(newk == INPUT_HANDLER) {
					input_handle = newk;
					// string pipe_cmd = concat(delrange(argv, 0, 4), " ");
					// echo("[_input] handle established as: " + pipe_cmd);
					llSleep(1);
					effector_restrict("input_handle", "redirchat:" + (string)(user_channel) + "=?,rediremote:" + (string)(user_channel) + "=?,sendchat=?");
					#ifdef DEBUG
					echo("[_input] redirect enabled");
					#endif
					
					// don't unqueue start-up jobs until now
					// as we won't be able to handle them
					#define OVERRIDE_READY
					system(SIGNAL_READY, PROGRAM_NAME);
					// task_end(INPUT_HANDLER);
				}
				return;
			} else {
				echo("Unknown notify directive " + action);
			}
			
			ci += 2;
			//echo("took notify " + action);
		}
		
		integer can_hear = allow_recvchat && (!vox_input) && hearing_state;
		if(!can_hear && !hearing_captured) {
			hearing_captured = TRUE;
			pipe_open(["p:" + (string)VOX_LISTENER + " n:" + (string)avatar + " from 0 " + NULL_KEY + " " + PROGRAM_NAME]);
			#ifdef DEBUG
				echo("[_input] captured hearing: "
					+ (string)allow_recvchat + ", "
					+ (string)vox_input + ", "
					+ (string)hearing_state
				);
			#endif
		} else if(can_hear && hearing_captured) {
			hearing_captured = FALSE;
			pipe_close([VOX_LISTENER]);
			#ifdef DEBUG
				echo("[_input] released hearing: "
					+ (string)allow_recvchat + ", "
					+ (string)vox_input + ", "
					+ (string)hearing_state
				);
			#endif
		}
	} else if(n == SIGNAL_VOX) {
		// special reflective thing
		rvq_message = m;
		rvq_flag = VO_ADD;
		jump rvq;
	} else if(n == SIGNAL_UNKNOWN_SCRIPT) {
		echo("[_input] unknown program: " + m);
	}
	
	jump main_end;
	@rvq;
	
	#ifdef RVQ_DEBUG
	echo("/me :::: RVQ OP " + (string)rvq_flag + ": " + rvq_message);
	#endif
	if(rvq_flag == VO_ADD) {
		vo_queue += rvq_message;
	} else if(rvq_flag == VI_ADD) {
		vi_queue += [rvq_source, rvq_message];
	} else if(rvq_flag == VO_ALTER) {
		if(~vo_next_cell) {
			vo_cells = alter(vo_cells, [rvq_message], vo_next_cell, vo_next_cell);
			#ifdef RVQ_DEBUG
			echo("/me :::: RVQ AMEND VO DONE");
			#endif
		}
		#ifdef RVQ_DEBUG
		else echo("/me :::: RVQ NOTHING TO AMEND VO");
		#endif
		
		vo_next_cell += 2;
		if(vo_next_cell >= count(vo_cells)) {
			#ifdef RVQ_DEBUG
			echo("/me :::: RVQ EMIT VO");
			#endif
			// emit!
			string omsg = concat(vo_cells, "\"");
			
			print(output_pipe, avatar, omsg);
			
			// echo((string)voci + ", " + (string)vo_next_cell);
			
			if(vo_next_cell > 1)
				play_speech_marker();
			
			vo_cells = [];
			vo_next_cell = NOWHERE;
		} else {
			string cellmsg = gets(vo_cells, vo_next_cell);
			if(!(voice_state && mind_state) || (impedence_out > 0) || (impedence_mind > 0)) {
				// todo: bypass mind state for preset messages
				cellmsg = "...";
			}
			
			if(vox_output) {
				print(vo_intake, avatar, cellmsg);
			} else {
				rvq_message = cellmsg;
				rvq_flag = VO_ALTER;
				jump rvq;
			}
		}
	} else if(rvq_flag == VI_ALTER) {
		if(~vi_next_cell) {
			vi_cells = alter(vi_cells, [rvq_message], vi_next_cell, vi_next_cell);
		}
		
		vi_next_cell += 2;
		if(vi_next_cell >= count(vi_cells)) {
			// emit!
			string imsg = concat(vi_cells, "\"");
			string visn = vi_name(vi_source);
			
			llSetObjectName(visn);
			echo(imsg);
			llSetObjectName("");
			
			vi_cells = [];
			vi_next_cell = NOWHERE;
		} else {
			string cellmsg = gets(vi_cells, vi_next_cell);
			if((!hearing_state) || (impedence_in > 0)) {
				cellmsg = "...";
			}
			if(vox_input) {
				print(vi_intake, vi_source, cellmsg);
			} else {
				rvq_message = cellmsg;
				rvq_flag = VI_ALTER;
				jump rvq;
			}
		}
	}
	
	integer direction = 0;
	string new_message;
	
	// advance queue:
	if((rvq_flag & VI_ADD || rvq_flag == 0)
	&& !~vi_next_cell
	&& count(vi_queue)) { // input queue
	
		direction = VI_ADD;
		new_message = gets(vi_queue, 1);
		vi_source = gets(vi_queue, 0);
		vi_queue = delrange(vi_queue, 0, 1);
		
	} else if((rvq_flag < VI_ADD || rvq_flag == 0)
	&& !~vo_next_cell
	&& count(vo_queue)) { // output queue
	
		direction = VO_ADD;
		new_message = gets(vo_queue, 0);
		vo_queue = delitem(vo_queue, 0);
		#ifdef RVQ_DEBUG
		echo("/me :::: RVQ ADVANCE");
		#endif
	}
	#ifdef RVQ_DEBUG
	else
		echo("/me :::: RVQ NO ADVANCE; waiting on cells: " + concat(vo_cells, " ") + " (vo_next_cell=" + (string)vo_next_cell + ")");
	#endif
	
	if(direction) {
		list new_cells;
		integer nc = 0;
		integer ncc;
	
		if(llOrd(new_message, 0) == 0x2f) { // '/'
			new_cells = splitnulls(new_message, "\"");
			nc = 1;
			ncc = count(new_cells);
		} else {
			new_cells = [new_message];
			nc = 0;
			ncc = 0;
		}
		
		if(direction == VO_ADD) {
			vo_cells = new_cells;
			
			if(ncc == 1) {
				// nothing to process!
				vo_next_cell = NOWHERE;
				#ifdef RVQ_DEBUG
				echo("/me :::: RVQ REFLECT");
				#endif
				rvq_flag = VO_ALTER;
				rvq_message = rvq_source = "";
				jump rvq;
				// run_vox_queue(VO_ALTER, "", "");
			} else {
				#ifdef RVQ_DEBUG
				echo("/me :::: RVQ SEEK");
				#endif
				vo_next_cell = nc;
				string cellmsg = gets(new_cells, nc);
				if(!(voice_state && mind_state) || (impedence_out > 0) || (impedence_mind > 0)) {
					// todo: bypass mind state for preset messages
					cellmsg = "...";
				}
				if(vox_output) {
					print(vo_intake, avatar, cellmsg);
				} else {
					rvq_message = cellmsg;
					rvq_flag = VO_ALTER;
					jump rvq;
				}
			}
		} else { // VI_ADD
			vi_cells = new_cells;
			
			if(ncc == 1) {
				// nothing to process!
				vi_next_cell = NOWHERE;
				rvq_flag = VI_ALTER;
				rvq_message = rvq_source = "";
				jump rvq;
				// run_vox_queue(VI_ALTER, "", "");
			} else {
				vi_next_cell = nc;
				string cellmsg = gets(new_cells, nc);
				if((!hearing_state) || (impedence_in > 0)) {
					cellmsg = "...";
				}
				if(vox_input) {
					print(vi_intake, vi_source, cellmsg);
				} else {
					rvq_message = cellmsg;
					rvq_flag = VI_ALTER;
					jump rvq;
				}
			}
		}
	}
	
	@main_end;
}

#define EXT_EVENT_HANDLER "ARES/system/input.event.lsl"
#include <ARES/program>
