/* =========================================================================
 *
 *             Nanite Systems Applied Research Execution System
 *  
 *            Copyright (c) 2022–2023 Nanite Systems Corporation
 *  
 * =========================================================================
 *
 *  Menus System Module
 *
 *  This program is covered under the terms of the ARES Software Copyright
 *  License, Section 4 (ASCL-iv). It is a closed-source, proprietary,
 *  commercial product and may not be distributed.
 *  
 *  To see the full text of the ASCL, type 'help license' on any standard
 *  ARES distribution, or visit http://nanite-systems.com/ASCL for the
 *  current version.
 *
 *  DISCLAIMER
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS
 *  IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 *  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 *  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 *  DAMAGES HOWEVER CAUSED ON ANY THEORY OF LIABILITY ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGE.
 *
 * =========================================================================
 *
 */

#include <ARES/a>
#include <ARES/api/auth.h.lsl>
#define CLIENT_VERSION ARES_VERSION
#define CLIENT_VERSION_TAGS ARES_VERSION_TAGS

// devices channel:
integer C_LIGHT_BUS;

#define SESSION_LOCATION 0
#define SESSION_USER 1
#define SESSION_TIMEOUT 2
#define SESSION_TOPIC 3
#define SESSION_PAGE 4
#define SESSION_REFRESHING 5
#define SESSION_PROTOCOL 6
#define SESSION_CHANNEL 7

#define PROTOCOL_LIGHT 0
#define PROTOCOL_A2A 1
#define PROTOCOL_PRIVATE 2

#define BUTTON_LABEL 0
#define BUTTON_TYPE 1
#define BUTTON_ACTION 2
#define BUTTON_DATA 3

#define BUTTON_TYPE_BASIC 0
#define BUTTON_TYPE_GO 1
#define BUTTON_TYPE_BOOLEAN 2
#define BUTTON_TYPE_STRING 3
#define BUTTON_TYPE_NEXT -1
#define BUTTON_TYPE_PREVIOUS -2
#define BUTTON_TYPE_FILE 4

#define BACK_BUTTON_LABEL "back"
#define REFRESH_BUTTON_LABEL "refresh"

list waiting_users; // avatars we are rezzing menu HUDs for

string sessions; // device:[location,user,timeout,topic,page,refreshing?,protocol,channel]

key rlv_check_pipe;

string footer;
string header;

// page -1 returns all buttons
list get_buttons(string menu_def, integer page) {
	list buttons;
	string dtype = llJsonValueType(menu_def, ["d"]); // dynamic buttons
	if(dtype == JSON_ARRAY)
		buttons += js2list(getjs(menu_def, ["d"]));
	else if(dtype == JSON_OBJECT) {
		integer dmax = count(jskeys(getjs(menu_def, ["d"])));
		integer di = 0;
		while(di < dmax)
			buttons += getjs(menu_def, ["d", di++]);
	}
	
	string back_button;
	
	string etype = llJsonValueType(menu_def, ["e"]); // static buttons
	if(etype == JSON_ARRAY) {
		list ebuttons = js2list(getjs(menu_def, ["e"]));
		integer emax = count(ebuttons);
		integer ei = 0;
		while(ei < emax) {
			string b = gets(ebuttons, ei++);
			if(getjs(b, [0]) == BACK_BUTTON_LABEL && getjs(b, [1]) == "1") {
				back_button = b;
			} else {
				buttons += b;
			}
		}
	} else if(etype == JSON_OBJECT) {
		integer emax = count(jskeys(getjs(menu_def, ["e"])));
		integer ei = 0;
		while(ei < emax) {
			string b = getjs(menu_def, ["e", ei++]);
			if(getjs(b, [0]) == BACK_BUTTON_LABEL && getjs(b, [1]) == "1") {
				back_button = b;
			} else {
				buttons += b;
			}
		}
	}
	
	if(count(buttons) > 23 && page != NOWHERE && back_button != "") {
		buttons = sublist(buttons, page * 21, (page + 1) * 21);
		while(count(buttons) < 21)
			buttons += "[\"\",0,\"\"]";
		buttons = sublist(buttons, 0, 6)
			+ ["[\"< PREVIOUS\",-2]"]
			+ sublist(buttons, 7, 13)
			+ back_button
			+ sublist(buttons, 14, 20)
			+ "[\"NEXT >\",-1]";
	} else {
		buttons += back_button;
	}
	
	return buttons;
}

#define menu_message(id, msg) { if(id == DAEMON) {\
		e_call(C_VARIATYPE, E_SIGNAL_CALL,\
			(string)avatar + " " + (string)avatar + " variatype " + msg\
		);\
	} else {\
		io_tell(id, getjs(sessions, [id, SESSION_CHANNEL]), msg);\
	} }

show_menu(key id, string menu_name, integer page) {
	string menu_def = llLinksetDataRead("m:" + menu_name);
	
	// echo("menu to show: " + menu_def);
	
	list button_text;
		
	list buttons = get_buttons(menu_def, page);
	// echo("Buttons found: " + concat(buttons, ", "));
	
	integer bmax = count(buttons);
	integer bi = 0;
	while(bi < bmax) {
		string button = gets(buttons, bi);
		if(button != "") {
			string button_label = getjs(button, [BUTTON_LABEL]);
			integer button_type = (integer)getjs(button, [BUTTON_TYPE]);
			// (BUTTON_ACTION)
			string button_data_source = getjs(button, [BUTTON_DATA]);
			
			if(button_type == BUTTON_TYPE_FILE || button_type < BUTTON_TYPE_BOOLEAN) {
				button_text += button_label;
			} else if(button_type >= BUTTON_TYPE_BOOLEAN) {
				integer mmi = NOWHERE;
				integer maskmode = strpos(button_data_source, "|");
				if(~maskmode) {
					mmi = (integer)delstring(button_data_source, 0, maskmode);
					button_data_source = delstring(button_data_source, maskmode, LAST);
					// echo("mask mode: bit " + (string)mmi + " of " + button_data_source);
				}
				list bds = splitnulls(button_data_source, ".");
				string s = getdbl(gets(bds, 0), delitem(bds, 0));
				if(button_type == BUTTON_TYPE_BOOLEAN) {
					integer si = (integer)s;
					
					if(~mmi)
						si = (si & (1 << mmi)) && TRUE;
					
					s = gets(["off", "on", "???"], si);
				}
				button_text += button_label + ": " + s;
			}
		}
		
		++bi;
	}
	
	string desc = getjs(menu_def, ["desc"]);
	if(~strpos(desc, "$topic")) {
		string topic = getjs(sessions, [(string)id, SESSION_TOPIC]);
		if(topic == "")
			topic = "(???)";
			
		desc = replace(desc, "$topic", topic);
	}
	
	string msg = "menu " + menu_name
		+ "\n" + header
		+ "\n" + footer
		+ "\n" + desc
		+ "\n" + concat(button_text, "\n");
	
	menu_message(id, msg);
	
	sessions = setjs(sessions, [(string)id, SESSION_TIMEOUT], (string)(llGetUnixTime() + 30));
}

main(integer src, integer n, string m, key outs, key ins, key user) {
	@internal_restart;
	if(n == SIGNAL_INVOKE) {
		list argv = split(m, " ");
		integer argc = count(argv);
		string action = gets(argv, 1);
		
		string msg;
		
		// REMINDER NOT TO MUTATE session, ALTER sessions DIRECTLY
		
		// session expiry code:
		list session_keys = jskeys(sessions);
		integer sxi = count(session_keys);
		while(sxi--) {
			key sx = getk(session_keys, sxi);
			if(sx != JSON_INVALID) {
				if(llGetOwnerKey(sx) == sx
				|| (integer)getjs(sessions, [sx, SESSION_TIMEOUT]) < llGetUnixTime()) {
					menu_message(sx, "menu end");
					sessions = setjs(sessions, [sx], JSON_DELETE);
					#ifdef DEBUG
					echo("[_menu] "+"retiring session " + (string)sx);
					#endif
				}
			}
		}
		
		if(action == "solicit") {
			// if the user has an entry in menu.invite, then delete that and start a session
			// a menu.invite entry is only created after checking for the menu permission
			
			string menu_invites = getdbl("menu", ["invite"]);
			
			integer invite_found = 0;
			integer invite_deadline = (integer)getjs(menu_invites, [user]);
			if(invite_deadline >= llGetUnixTime()) {
				invite_found = 1;
				menu_invites = setjs(menu_invites, [user], JSON_DELETE);
				integer mi = index(waiting_users, user);
				if(~mi)
					waiting_users = delitem(waiting_users, mi);
			}
			
			// do cleanup:
			integer mutated = 0;
			if(menu_invites != "{}") {
				list invites = jskeys(menu_invites);
				integer i = count(invites);
				while(i--) {
					string k = gets(invites, i);
					integer deadline = (integer)getjs(menu_invites, [k]);
					if(deadline < llGetUnixTime()) {
						menu_invites = setjs(menu_invites, [k], JSON_DELETE);
						mutated = 1;
					}
					
					integer mi = index(waiting_users, (key)k);
					if(~mi)
						waiting_users = delitem(waiting_users, mi);
				}
			}
			
			if(mutated || invite_found)
				setdbl("menu", ["invite"], menu_invites);
			
			if(invite_found) {
				m = "* start " + gets(argv, 2) + " " + gets(argv, 3);
				#ifdef DEBUG
				echo("[_menu] "+"found invite; starting session");
				#endif
				jump internal_restart;
			}
			
		} else if(action == "build") {
			// string s_fs = llLinksetDataRead("fs");
			string category = gets(argv, 2);
			string topic = getjs(sessions, [(string)outs, SESSION_TOPIC]);
			string menu_name;
			string buttons;
			
			if(category == "fs") {
				list sources = jskeys(getdbl("fs", ["source"]));
				list views = jskeys(getdbl("fs", ["view"]));
				integer pi;
				if(~(pi = index(views, "parc")))
					views = delitem(views, pi);
					
				integer vmax = count(views);
				integer vi;
				while(vi < vmax) {
					string view = gets(views, vi);
					buttons = setjs(buttons, [JSON_APPEND], jsarray([
						view + " files...",
						1,
						"m:fs-view",
						view
					]));
					++vi;
				}
				
				vmax = count(sources);
				vi = 0;
				while(vi < vmax) {
					string source = gets(sources, vi);
					buttons = setjs(buttons, [JSON_APPEND], jsarray([
						source + ":>",
						1,
						"m:fs-source",
						source
					]));
					++vi;
				}
				
				menu_name = "m:fs";
			} else { // including 'source' and 'view'
				if(category == "view" || category == "source") {
					menu_name = "m:fs-" + category;
					category = topic;
				} else {
					menu_name = getdbl("fs", ["view", category, "menu"]);
				}
				
				if(menu_name != JSON_INVALID) {
					setdbl(menu_name, ["d"], "[]");
					
					list files = js2list(llLinksetDataRead("fs:" + category));
					integer fmax = count(files);
					integer fi = 0;
					while(fi < fmax) {
						string fn = gets(files, fi);
						buttons = setjs(buttons, [JSON_APPEND], jsarray([
							fn,
							4
						]));
						++fi;
					}
				} else {
					echo("[_menu] "+"no .menu definition for filesystem view '" + category + "'");
					return;
				}
			}
			
			// s_fs = "";
			setdbl(menu_name, ["d"], buttons);
			
		} else if(action == "prompt") {
			string cmd = concat(delrange(argv, 0, 1), " ");
			e_call(C_BASEBAND, E_SIGNAL_CALL,
				(string)outs + " " + (string)user + " baseband prompt " + jsobject([
					"message", "Specify parameters for the following command:\n\n\t" + cmd
					+ "\n\nhelp: http://support.nanite-systems.com/?id=2832#" + gets(argv, 2),
					"ins", ins,
					"command", cmd
				])
			);
		} else if(action == "select") {
			string session = getjs(sessions, [(string)outs]);
			if(session != JSON_INVALID) {
				// _menu select <button_number> <menu_name> <button_label>
				
				string menu_name = gets(argv, 3);
				string button_label = concat(delrange(argv, 0, 3), " ");
				
				string session_location = getjs(session, [SESSION_LOCATION]);
				
				string menu_def = llLinksetDataRead("m:" + session_location);
				integer page = (integer)getjs(session, [SESSION_PAGE]);
				
				if(menu_name != session_location) {
					echo("[_menu] "+"session location mismatch (in " + session_location + ", not " + menu_name + "); may be clicking too fast");
					show_menu(outs, session_location, page);
					return;
				}
				
				list buttons = get_buttons(menu_def, page);
				integer button_index = (integer)gets(argv, 2);
				if(button_index >= count(buttons))
					return;
				string button = gets(buttons, button_index);
				if(button == JSON_INVALID)
					return;
				
				string real_button_label = getjs(button, [BUTTON_LABEL]);
				#ifdef DEBUG
				if(button_label != real_button_label)
					echo("[_menu] "+"warning: button label mismatch; user pressed " + button_label + " but button actually reads " + real_button_label);
				#endif
				
				// string button = getdbl("m:" + session_location, ["e", ]);
				// echo("[_menu] "+"selected button: " + button + " from menu " + session_location);
				integer button_type = (integer)getjs(button, [BUTTON_TYPE]);
				string command = getjs(button, [BUTTON_ACTION]);
				
				string target_menu;
				string target_menu_permission;
				if(button_type == BUTTON_TYPE_GO) {
					target_menu = command;
				} else {
					target_menu = "m:" + session_location;
				}
				
				target_menu_permission = getdbl(target_menu, ["p"]);
				
				if(target_menu_permission == JSON_INVALID) {
					tell(user, 0, "Menu " + target_menu + " doesn't exist.");
				}
				
				#define A_GO 1
				#define A_ACT 2
				#define A_DENY 3
				integer action_result;
				
				sessions = setjs(sessions, [(string)outs, SESSION_USER], user);
				
				sessions = setjs(sessions, [(string)outs, SESSION_TIMEOUT], (string)(llGetUnixTime() + 30));
				
				integer sec_result = sec_check(user, target_menu_permission, outs, "", m);
				if(sec_result == DENIED) {
					tell(user, 0, "You are not authorized to access that menu.");
					action_result = A_DENY;
					if(real_button_label == BACK_BUTTON_LABEL) {
						msg = "Returning to main menu.";
						sessions = setjs(sessions, [(string)outs, SESSION_LOCATION], "main");
						sessions = setjs(sessions, [(string)outs, SESSION_PAGE], "0");
						
						show_menu(outs, "main", 0);
					}
				} else if(sec_result == ALLOWED) {
					if(button_type != BUTTON_TYPE_GO
					&& button_type != BUTTON_TYPE_NEXT
					&& button_type != BUTTON_TYPE_PREVIOUS) {
						string b_e0 = getjs(menu_def, ["e", "0"]);
						if(b_e0 == JSON_INVALID)
							b_e0 = getjs(menu_def, ["e", 0]);
						
						action_result = A_ACT;
						if(button_type == BUTTON_TYPE_FILE) {
							list parts = splitnulls(real_button_label, ".");
							string ext;
							string verb;
							if(count(parts) > 1) {
								ext = gets(parts, LAST);
								verb = getdbl("fs", ["open", ext]);
								
								command = verb + " " + real_button_label;
							}
							
							if(ext == "" || verb == JSON_INVALID) {
								print(user, user, "No file association: " + real_button_label);
							}
						}
							
						if(command != "") {
							if(~strpos(command, "$topic")) {
								string topic = getjs(sessions, [(string)outs, SESSION_TOPIC]);
								if(topic == "" || topic == JSON_INVALID) {
									msg = "Cannot execute command '" + command + "'; no topic set.";
									if(outs == DAEMON) {
										interface_sound("deny");
										echo(msg);
										return;
									}
									jump fail;
								} else {
									command = replace(command, "$topic", topic);
								}
							}
							invoke(command, user, outs, user);
						}
						
						// ideally we'd wait for a DONE message before doing this:
						
						if(getjs(b_e0, [BUTTON_LABEL]) == REFRESH_BUTTON_LABEL) {
							string b_e0_command = getjs(b_e0, [BUTTON_ACTION]);
							if(b_e0_command != command) {
								#ifdef DEBUG
								echo("[_menu] "+"now refreshing menu with command: " + b_e0_command);
								#endif
								list bargv = split(b_e0_command, " ");
								if(gets(bargv, 0) == "menu" || gets(bargv, 0) == "_menu") {
									main(src, SIGNAL_INVOKE, b_e0_command, outs, ins, user);
									/*
									if(gets(bargv, 1) == "build") {
										generate_view_menu(outs, gets(bargv, 2));
									} else {
										
									}*/
								} else {
									invoke(b_e0_command, user, outs, user);
								}
							}
						}
					} else {
						string target_menu_def = llLinksetDataRead(target_menu);
						
						action_result = A_GO;
					
						if(button_type == BUTTON_TYPE_NEXT || button_type == BUTTON_TYPE_PREVIOUS) {
							integer button_count = count(get_buttons(target_menu_def, NOWHERE));
							if(button_type == BUTTON_TYPE_NEXT)
								++page;
							else
								--page;
							
							integer page_count = llCeil((float)button_count / 21.0);
							if(page < 0)
								page = page_count - 1;
							else if(page >= page_count)
								page = 0;
							
							command = session_location;
							sessions = setjs(sessions, [(string)outs, SESSION_PAGE], (string)page);
						} else { // button type go
							command = delstring(command, 0, 1); // remove 'm:'
							sessions = setjs(sessions, [(string)outs, SESSION_LOCATION], command);
							sessions = setjs(sessions, [(string)outs, SESSION_PAGE], "0");
							
							string topic = getjs(button, [BUTTON_DATA]);
							if(topic == JSON_INVALID)
								topic = "";
							sessions = setjs(sessions, [(string)outs, SESSION_TOPIC], topic);
						}
						
						string b_e0 = getjs(target_menu_def, ["e", "0"]);
						if(b_e0 == JSON_INVALID)
							b_e0 = getjs(target_menu_def, ["e", 0]);
						
						if(getjs(b_e0, [BUTTON_LABEL]) == REFRESH_BUTTON_LABEL) {
							string b_e0_command = getjs(b_e0, [BUTTON_ACTION]);
							
							#ifdef DEBUG
							echo("[_menu] "+"now refreshing menu with command: " + b_e0_command);
							#endif
							list bargv = split(b_e0_command, " ");
							if(gets(bargv, 0) == "menu" || gets(bargv, 0) == "_menu") {
								main(src, SIGNAL_INVOKE, b_e0_command, outs, ins, user);
								/*
								if(gets(bargv, 1) == "build") {
									generate_view_menu(outs, gets(bargv, 2));
								} else {
									main(src, SIGNAL_INVOKE, b_e0_command, outs, ins, user);
								}*/
							} else {
								invoke(b_e0_command, user, outs, user);
							}
						}
						
						show_menu(outs, command, page);
					}
				}
				// show_menu(outs, );
				
				if(outs == DAEMON && action_result) {
					string sound_name = gets(["", "go", "act", "deny"], action_result);
					interface_sound(sound_name);
				}
				
				setdbl("menu", ["session"], session);
			} else {
				msg = "Your menu session has expired.";
			}
			
		} else if(action == "start") {
			#ifdef DEBUG
			echo("[_menu] "+"session start command: " + m);
			#endif
			// menu start [<location> [<channel (or 0 for light bus)> [<topic>]]]
			string session = getjs(sessions, [(string)outs]);
			string location = gets(argv, 2);
			integer page;
			if(session != JSON_INVALID) {
				if(location == "")
					location = getjs(sessions, [(string)outs, SESSION_LOCATION]);
				else
					sessions = setjs(sessions, [(string)outs, SESSION_LOCATION], location);
				
				// allow switching channels and protocols:
				if(argc > 3) {
					integer newc = (integer)gets(argv, 3);
					if(newc != 0) {
						sessions = setjs(sessions, [(string)outs, SESSION_PROTOCOL], (string)PROTOCOL_PRIVATE);
						sessions = setjs(sessions, [(string)outs, SESSION_CHANNEL], (string)newc);
					} else {
						sessions = setjs(sessions, [(string)outs, SESSION_PROTOCOL], (string)PROTOCOL_LIGHT);
						sessions = setjs(sessions, [(string)outs, SESSION_CHANNEL], (string)C_LIGHT_BUS);
					}
				}
				
				// allow switching topics:
				if(argc > 4)
					sessions = setjs(sessions, [(string)outs, SESSION_TOPIC], gets(argv, 4));
				
				page = (integer)getjs(sessions, [(string)outs, SESSION_PAGE]);
				#ifdef DEBUG
				echo("[_menu] "+"device " + llKey2Name(outs) + " already has a session");
				#endif
			} else {
				if(location == "")
					location = "main";
				
				integer protocol = PROTOCOL_LIGHT;
				integer c = C_LIGHT_BUS;
				if(argc > 3) {
					integer newc = (integer)gets(argv, 3);
					if(newc != 0) {
						protocol = PROTOCOL_PRIVATE;
						c = newc;
					}
				}
				
				string topic;
				if(argc > 4) {
					topic = gets(argv, 4);
				}
				
				sessions = setjs(sessions, [(string)outs],
					jsarray([location, user, llGetUnixTime() + 30, topic, 0, FALSE, protocol, c]));
				
				#ifdef DEBUG
				echo("[_menu] "+"started new session for device " + llKey2Name(outs));
				#endif
			}
			
			show_menu(outs, location, page);
			setdbl("menu", ["session"], sessions);
		
		} else if(action == "end") {
			string session = getjs(sessions, [(string)outs]);
			if(session != JSON_INVALID) {
				menu_message(outs, "menu end");
				sessions = setjs(sessions, [(string)outs], JSON_DELETE);
				setdbl("menu", ["session"], sessions);
			}
		} else if(action == "request") {
			// someone touched the hardware; may they use menus?
			vector where = getv(geto(user, [OBJECT_POS]), 0);
			integer perm;
			if(llVecDist(where, llGetPos()) >= 20) {
				msg = "You are too far away to access " + llGetLinkName(R_DAEMON) + ".";
			} else if((integer)getdbl("status", ["on"]) == 0) {
				invoke("_power on", outs, NULL_KEY, user);
			} else if((perm = sec_check(user, "menu", user, "", m)) == ALLOWED) {
				if(!contains(waiting_users, user)) {
					setdbl("menu", ["invite", user], (string)(llGetUnixTime() + 30));
					integer pflags = llGetParcelFlags(llGetPos());
					list ptraits = llGetParcelDetails(llGetPos(), [
						PARCEL_DETAILS_GROUP,
						PARCEL_DETAILS_OWNER,
						PARCEL_DETAILS_PRIM_CAPACITY,
						PARCEL_DETAILS_PRIM_USED
					]);
					key pgroup = getk(ptraits, 0);
					key powner = getk(ptraits, 1);
					integer pfree = geti(ptraits, 2) - geti(ptraits, 3);
					if((pflags & PARCEL_FLAG_ALLOW_CREATE_OBJECTS
					|| (pflags & PARCEL_FLAG_ALLOW_CREATE_GROUP_OBJECTS && llSameGroup(pgroup))
					|| (avatar == powner)) && pfree >= 14) {
						waiting_users += user;
						// check pfree here and offer heavier HUDs if high enough
						llRezObject("_AMLite", llGetPos(), ZV, ZR, 1);
					} else {
						msg = "To access " + llGetLinkName(R_DAEMON) + "'s menus, please wear the AMLite HUD.";
						llGiveInventory(user, "_AMLite");
					}
				}
			} else if(perm == DENIED) {
				msg = "You do not have permission to access " + llGetLinkName(R_DAEMON) + "'s menus.";
				if(user != avatar)
					echo("Denied menu access to secondlife:///app/agent/" + (string)user +"/about");
			}
		
		} else if(argc == 1) {
			// 'outs' is usually a device that can't receive messages,
			// so only when the syntax unambiguously indicates manual input
			// do we trust it:
			
			print(outs, user, "[" + PROGRAM_NAME + "] control software for ARES menu interfaces (" + (string)llGetUsedMemory() + " bytes used)\nsee 'help menu' for usage and the bundled menu.db file for more information");
		}
		
		@fail;
		if(msg != "") {
			print(user, user, msg);
		}
		
	} else if(n == SIGNAL_DONE) {
		ins = substr(m, 0, 35);
		string session = getjs(sessions, [(string)ins]);
		if(session != JSON_INVALID) {
			show_menu(ins,
				getjs(session, [SESSION_LOCATION]),
				(integer)getjs(session, [SESSION_PAGE])
			);
			#ifdef DEBUG
			echo("[_menu] "+"got done for session: " + (string)ins);
			#endif
		} else {
			#ifdef DEBUG
			echo("[_menu] "+"got done for unknown session: " + (string)ins);
			#endif
		}
	} else if(n == SIGNAL_INIT) {
		C_LIGHT_BUS = 105 - (integer)("0x" + substr(avatar, 29, 35));
		#ifdef DEBUG
			echo("[" + PROGRAM_NAME + "] init event");
		#endif
		set_mode(MODE_ACCEPT_DONE);
		
		sessions = "{}";
		// setdbl("menu", ["session"], "{}");
		// sessions = getdbl("menu", ["session"]);
		
		header = "               Command Interface        version " + ARES_VERSION + " " + ARES_VERSION_TAGS;
		footer = "(c) " + ARES_YEAR + " Nanite Systems                " + getdbl("id", ["serial"]);
		hook_events([EVENT_ON_REZ, EVENT_REZ_OBJECT]);
	} else if(n == SIGNAL_EVENT) {
		if(m == (string)EVENT_ON_REZ) {
			setdbl("menu", ["session"], sessions = "{}");
		} else if((integer)substr(m, 0, 2) == EVENT_REZ_OBJECT) {
			key object = delstring(m, 0, 2);
			string object_name = llKey2Name(object);
			if(object_name == "_AMLite" || object_name == "_AMenu") {
				key u = getk(waiting_users, 0);
				waiting_users = delitem(waiting_users, 0);
				
				#define AM_CLIENT_ASSIGNMENT_CHANNEL 0x4D656E7F
				tell(object, AM_CLIENT_ASSIGNMENT_CHANNEL, u);
			}
		} else {
			echo("_menu: unexpected event " + m);
		}
	} else {
		echo("[" + PROGRAM_NAME + "] unimplemented signal " + (string)n + ": " + m);
	}
}

#include <ARES/program>
